/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.specific;

import ch.cern.lhcb.xmleditor.dtdparser.CustomDTDParser;
import ch.cern.lhcb.xmleditor.gui.MessageDisplayer;

import com.wutka.dtd.DTD;

/**
 * this encapsulates a dtd file. This class is able to find the dtd file
 * associated to a given Xml file and to parse it. It provides then an access to
 * a DTD object
 * 
 * @author Sebastien Ponce
 */
public class DtdFile extends RelativeFile {

    /**
     * The DTD associated with this file
     */
    private DTD dtd;

    /**
     * the MessageDisplayer associated to this file for error display This is
     * the first MessageDisplayer that opened the file
     */
    private MessageDisplayer msgDis;

    /**
     * constructor
     * 
     * @param parent a file used as the current position when looking for child
     * @param child the file we are looking for. This name is relative to the
     *        position of parent
     * @param msgDis the MessageDisplayer used for error display
     */
    public DtdFile(XmlFile parent, String child, MessageDisplayer msgDis) {
        super(parent.getParentFile(), child);
        this.msgDis = msgDis;
        dtd = null;
    }

    /**
     * constructor
     * 
     * @param fullFileName the name of the file to open
     * @param msgDis the MessageDisplayer used for error display
     */
    public DtdFile(String fullFileName, MessageDisplayer msgDis) {
        super(fullFileName);
        this.msgDis = msgDis;
        dtd = null;
    }

    /**
     * parses the file
     * 
     * @return true if the parsing was successfull
     */
    public boolean parse() {
        try {
            // creates a parser
            CustomDTDParser parser = new CustomDTDParser(this);
            // does the actual parsing and retrieves the document
            dtd = parser.parse();
            // assumes the parsing was successfull
            return true;
        } catch (java.io.IOException e) {
            msgDis.popupMessage("An error occured while parsing the dtd file "
                    + getAbsolutePath());
            msgDis.displayMessage("The file " + getAbsolutePath()
                    + " cannot be parsed. "
                    + "The following exception was raised :\n" + e + "\n");
            return false;
        }
    }

    /**
     * returns the dtd parsed from this file or null if the file was not parsed
     * 
     * @return the parsed dtd
     */
    public DTD getDtd() {
        return dtd;
    }

}