/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.specific;

import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import org.apache.xerces.dom.DocumentTypeImpl;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Attr;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import ch.cern.lhcb.xmleditor.Editor;
import ch.cern.lhcb.xmleditor.gui.MessageDisplayer;
import ch.cern.lhcb.xmleditor.xerces.CustomDOMParser;
import ch.cern.lhcb.xmleditor.xerces.DocumentImpl;
import ch.cern.lhcb.xmleditor.xerces.ErrorHandler;

/**
 * this encapsulates an Xml file. This class is able to find an Xml file and to
 * parse it. It provides then a DOM document representing the file
 * 
 * @author Sebastien Ponce
 */
public class XmlFile extends RelativeFile {

    /**
     * the document parsed from the file
     */
    private DocumentImpl document;

    /**
     * The XmlDisplay associated with this Xml file
     */
    private XmlDisplay xmlDisplay;

    /**
     * the MessageDisplayer associated to this file for error display This is
     * the first TreeView that opened the file
     */
    private MessageDisplayer msgDis;

    /**
     * the editor where this file is displayed
     */
    private Editor editor;

    /**
     * the serializer used to serialize
     */
    private XMLSerializer serializer;

    /**
     * the date of the last modification done on this file
     */
    public long lastModified;

    /**
     * says whether this file was modified by the user or not
     */
    public boolean upToDate;

    /**
     * says whether this file has a name or not. The only way a file could have
     * no name is when it was created by the File->New menu item
     */
    public final boolean hasName;

    /**
     * constructor
     * 
     * @param parent a file used as the current position when looking for child
     * @param child the file we are looking for. This name is relative to the
     *        position of parent
     * @param msgDis the MessageDisplayer used for error displaying
     * @param editor editor
     */
    public XmlFile(XmlFile parent, String child, MessageDisplayer msgDis,
            Editor editor) {
        super(parent.getParentFile(), child);
        this.msgDis = msgDis;
        this.editor = editor;
        document = null;
        serializer = null;
        upToDate = true;
        hasName = true;
        lastModified = 0;
    }

    /**
     * constructor
     * 
     * @param fullFileName the name of the file to open
     * @param msgDis the MessageDisplayer used for error displaying
     * @param editor editor
     */
    public XmlFile(String fullFileName, MessageDisplayer msgDis, Editor editor) {
        super(fullFileName);
        this.msgDis = msgDis;
        this.editor = editor;
        document = null;
        serializer = null;
        upToDate = true;
        hasName = true;
        lastModified = 0;
    }

    /**
     * constructor
     * 
     * @param name the (temporary) name of the file
     * @param docType the name of the document type
     * @param dtdName the name of the file containing the dtd
     * @param msgDis the MessageDisplayer used for error displaying
     * @param editor editor
     */
    public XmlFile(String name, String docType, String dtdName,
            MessageDisplayer msgDis, Editor editor) {
        super(name);
        this.msgDis = msgDis;
        this.editor = editor;
        serializer = null;
        upToDate = false;
        hasName = false;
        lastModified = 0;
        // creates a new document
        DocumentType documentType = new DocumentTypeImpl(null, null, null,
                dtdName);
        document = new DocumentImpl(documentType);
        document.setXmlFile(this);
        Element elem = document.createElement(docType);
        document.appendChild(elem);
        // sets the UndoManager for this document
        document.setUndoManager(editor.getUndoManager());
    }

    /**
     * constructor
     * 
     * @param oldFile an old file
     * @param fileName the new name for this old file
     * @param editor editor
     */
    public XmlFile(XmlFile oldFile, String fileName, Editor editor) {
        super(fileName);
        this.xmlDisplay = oldFile.getXmlDisplay();
        this.msgDis = oldFile.msgDis;
        this.serializer = null;
        this.upToDate = false;
        this.hasName = true;
        this.editor = editor;
        // in order to be able to change the name of the dtd for this file
        // (we need to make it relative to the new name of the file)
        // we have to recreate the document with a new doctype
        //
        // First, computes the new dtd name, to make it relative to the new file
        // name
        String dtdName = oldFile.getXmlDisplay().getDtdFile().getRelativePath(
                this);
        // creates a new document with the right dtd name
        DocumentType documentType = new DocumentTypeImpl(null, null, null,
                dtdName);
        document = new DocumentImpl(documentType);
        document.setXmlFile(this);
        // put into the new document the nodes of the old one
        while (oldFile.document.hasChildNodes()) {
            Node child = oldFile.document.getFirstChild();
            // do not adopt the documenttype
            if (child instanceof DocumentType) {
                oldFile.document.removeChild(child);
            } else {
                document.adoptNode(child);
                this.document.appendChild(child);
            }
        }
        // sets the UndoManager for this document
        document.setUndoManager(editor.getUndoManager());
    }

    /**
     * parses the file
     * 
     * @return true if parsing was successfull
     */
    public boolean parse() {
        // gets a parser
        CustomDOMParser parser = CustomDOMParser.getParser();
        // gets the corresponding ErrorHandler
        ErrorHandler errorHandler = (ErrorHandler) parser.getErrorHandler();
        // the returned value
        boolean result = true;
        // does the actual parsing and retrieves the document
        try {
            // just before parsing, notices the data of the last modification
            // of this file
            lastModified = this.lastModified();
            // now, parses
            parser.parse("File:///" + getAbsolutePath());
            document = (DocumentImpl) parser.getDocument();
            if (document == null) {
                msgDis
                        .popupMessage("The parsing did not provide a valid document.");
                result = false;
            } else {
                // associates this XmlFile object to the document
                document.setXmlFile(this);
                // sets the UndoManager for this document
                document.setUndoManager(editor.getUndoManager());
            }
        } catch (java.io.IOException e) {
            msgDis.popupMessage("Unable to open this file for parsing.");
            result = false;
        } catch (org.xml.sax.SAXException e) {
            msgDis.popupMessage("Unable to parse this file.");
            result = false;
        }
        // Displays parsing errors
        if (errorHandler.errorsExisting()) {
            msgDis.displayMessage("The parsing of file " + getAbsolutePath()
                    + " resulted in the following errors and warnings :\n");
            errorHandler.displayErrors(msgDis);
            msgDis.displayMessage("\n");
        }
        return result;
    }

    /**
     * this is provided for debug purposes. It prints the DOM tree corresponding
     * to this file.
     */
    public void dump() {
        dump(document);
    }

    /**
     * this is provided for debug purposes. It prints the DOM tree corresponding
     * to the Node node.
     * 
     * @param node the top of the subtree to be displayed
     */
    public static void dump(Node node) {
        dump(node, 0);
    }

    /**
     * this displays a subDOMtree from node node, with an indentation of depth
     * 
     * @param node the top of the subtree
     * @param depth the depth of the top of the subtree, to manage indentation
     */
    private static void dump(Node node, int depth) {
        // this should be useless but it seems that some attributes are created
        // by the documents with null children... To be investigated, there
        // should
        // be a bug in xerces
        if (node == null) { return; }
        // indent
        for (int i = 0; i < depth; i++)
            System.out.print("    ");
        if (node instanceof Attr) {
            System.out.println(node.toString() + " owned by "
                    + ((Attr) node).getOwnerElement());
        } else {
            System.out.println(node.toString());
        }
        NamedNodeMap attributes = node.getAttributes();
        if (attributes != null) {
            if (attributes.getLength() > 0) {
                for (int i = 0; i < depth; i++)
                    System.out.print("    ");
                System.out.println("  Attributes :");
            }
            for (int i = 0; i < attributes.getLength(); i++)
                dump(attributes.item(i), depth + 1);
        }
        NodeList children = node.getChildNodes();
        if (children != null) {
            if (children.getLength() > 0) {
                for (int i = 0; i < depth; i++)
                    System.out.print("    ");
                System.out.println("  Children :");
            }
            for (int i = 0; i < children.getLength(); i++)
                dump(children.item(i), depth + 1);
        }
    }

    /**
     * writes the document coming from this file (and potentially modified) to
     * the OutputStream outputStream
     * 
     * @param outputStream the stream where to write
     * @throws IOException if the serialization crashes
     */
    private void writeFile(OutputStream outputStream) throws IOException {
        // check if a serializer exists
        if (serializer != null) {
            // if yes, resets it
            serializer.reset();
            serializer.setOutputByteStream(outputStream);
        } else {
            // else creates one
            OutputFormat outputFormat = new OutputFormat(document);
            // in order to make indenting work, we cannot preserve spaces
            outputFormat.setPreserveSpace(false);
            outputFormat.setIndenting(true);
            outputFormat.setLineWidth(0);
            serializer = new XMLSerializer(outputStream, outputFormat);
        }
        // serializes the document
        serializer.serialize(document);
    }

    /**
     * returns the content of the document coming from this file (potentially
     * modified)
     * 
     * @return the content of the document coming from this file
     * @throws IOException 
     */
    public String getContent() throws IOException {
        // creates an output stream
        OutputStream outputStream = new ByteArrayOutputStream();
        // writes the current document to it
        writeFile(outputStream);
        // return the content of the OutputStream
        return outputStream.toString();
    }

    /**
     * tries to save this file using the parsed document that was potentially
     * modified if not successfull, lets the user know and returns false
     * 
     * @return true if successfull
     */
    public boolean save() {
        // if this file was just created, asks the user for a name
        if (!hasName) {
            // this file is not named, it cannot be saved
            msgDis.popupMessage("Unable to save file : no name.");
            msgDis
                    .displayMessage("An error occured while saving "
                            + this.getName()
                            + "\n"
                            + "This file has no name. This message should never be displayed, "
                            + "please, report this problem to the gaudi team.");
            return false;
        }
        try {
            // creates an output stream, using the current file
            OutputStream outputStream = new FileOutputStream(this);
            // writes the current document to the file
            writeFile(outputStream);
            // closes the open file
            outputStream.close();
            // updates its state
            this.upToDate = true;
            lastModified = this.lastModified();
            // returns
            return true;
        } catch (IOException ioException) {
            msgDis.popupMessage("Unexpected crash during saving.");
            msgDis.displayMessage("An error occured while saving "
                    + this.getName() + " : " + ioException + "\n"
                    + "You probably don't have the permission to write in "
                    + "this file. Use \"Save as\" to save your work.");
            return false;
        }
    }

    /**
     * returns the document parsed from the file or null if the file was not
     * parsed
     * 
     * @return the parsed document
     */
    public DocumentImpl getDocument() {
        return document;
    }

    /**
     * returns the XmlDispla associated with the file or null if the file was
     * not parsed
     * 
     * @return the XmlDisplay
     */
    public XmlDisplay getXmlDisplay() {
        return xmlDisplay;
    }

    /**
     * sets the XmlDisplay associated to this file
     * 
     * @param xmlDisplay the new XmlDisplay
     */
    public void setXmlDisplay(XmlDisplay xmlDisplay) {
        this.xmlDisplay = xmlDisplay;
    }

}