/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.undo;

import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;

import org.w3c.dom.Document;

import ch.cern.lhcb.xmleditor.Editor;

/**
 * this is an extension of the default UndoManager that handles some application
 * specific tasks
 */

public class UndoManager extends javax.swing.undo.UndoManager {

    /**
     * the editor using this UndoManager
     */
    protected Editor editor;

    /**
     * default constructor
     * 
     * @param editor the editor using this UndoManager
     */
    public UndoManager(Editor editor) {
        super();
        this.editor = editor;
    }

    /**
     * If inProgress, inserts anEdit and removes any old edits that were after.
     * Overridden here to update the editor menus accordingly.
     */
    public boolean addEdit(UndoableEdit anEdit) {
        boolean result = super.addEdit(anEdit);
        editor.updateEditMenu();
        return result;
    }

    /**
     * If this UndoManager is inProgress, undoes the last significant
     * UndoableEdit. Overridden here to update the editor menus accordingly.
     */
    public void undo() throws CannotUndoException {
        super.undo();
        editor.updateEditMenu();
    }

    /**
     * this undoes all actions in this UndoManager
     */
    public void undoAll() {
        while (this.canUndo()) {
            this.undo();
        }
    }

    /**
     * If this UndoManager is inProgress, redoes the last significant
     * UndoableEdit. Overridden here to update the editor menus accordingly.
     */
    public void redo() throws CannotRedoException {
        super.redo();
        editor.updateEditMenu();
    }

    /**
     * this redoes all actions in this UndoManager
     */
    public void redoAll() {
        while (this.canRedo()) {
            this.redo();
        }
    }

    /**
     * This ends the next significant edit to be redone in case it is a non
     * ended compound edit. This ensures that the next added edit will be at the
     * top level and not inside a compound edit.
     */
    public void endAnyInnerCompound() {
        UndoableEdit edit = this.editToBeUndone();
        if (edit instanceof javax.swing.undo.CompoundEdit) {
            ((javax.swing.undo.CompoundEdit) edit).end();
            editor.updateEditMenu();
        }
    }

    /**
     * accessor to the editor using this UndoManager
     * 
     * @return the editor using this UndoManager
     */
    public Editor getEditor() {
        return editor;
    }

    /**
     * Send die to each subedit, in the reverse of the order that they were
     * added. Then remove these edits from the UndoManager.
     * 
     * @param document a document
     */
    public void conditionnalDieAndTrim(Document document) {
        // Removes all unnecessary Edits
        int size = edits.size();
        for (int i = size - 1; i >= 0; i--) {
            if (edits.elementAt(i) instanceof DocumentUndoableEdit) {
                DocumentUndoableEdit e = (DocumentUndoableEdit) edits
                        .elementAt(i);
                if (e.conditionnalDieAndTrim(document)) {
                    this.trimEdits(i, i);
                }
            }
        }
        // update the Editor Menu
        editor.updateEditMenu();
    }

}