/*
 * com.simulogic.cese -- Cell Electrophysiology Simulation Environment (CESE)
 * 
 * Copyright (C) 1999 - 2005 Sergey Missan
 * 
 * Dalhousie Univ., Halifax, NS, Canada
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 *  
 */

package com.simulogic.cese;

import java.awt.Toolkit;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.beans.PropertyDescriptor;
import java.io.File;
import java.io.IOException;
import java.util.Vector;

import javax.help.CSH;
import javax.swing.Box;
import javax.swing.ButtonGroup;
import javax.swing.ComboBoxModel;
import javax.swing.ImageIcon;
import javax.swing.JComboBox;
import javax.swing.JMenu;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JTabbedPane;
import javax.swing.tree.TreePath;

import com.simulogic.cese.actions.OpenRecentParamsAction;
import com.simulogic.cese.properties.ExportPane;
import com.simulogic.cese.properties.PropertyPane;
import com.simulogic.utils.ExceptionDialog;
import com.simulogic.utils.RecentFileMenu;
import com.simulogic.utils.SwingWorker;

/**
 * This class aggregates all the actions related to the model management and
 * lifecycle
 */
public class ModelManager extends JPanel implements
		java.beans.PropertyChangeListener {

	private UserHome home;

	private SimulationOptions options;

	private SwingWorker worker;

	private RecordBuilder builder;

	private long time;

	private RecentFileMenu recentParameters;

	private ModelInfo modelInfo;

	private ExportPane exportPane;

	private PropertyPane propertyPane;

	private ExperimentTypeChanger experimentType;

	private javax.swing.JTabbedPane paramsTabs;

	private javax.swing.JComboBox modelSelector;

	private javax.swing.JSplitPane modelSplitPane;

	private javax.swing.JPanel paramsPanel;

	private javax.swing.JTextField statusBar;

	private String iteration = "";

	private static final CESE dm = CESE.getEnvironment();

	/** Default constructor */
	public ModelManager() {
		super();
		initComponents();
	}

	private void initComponents() {
		options = new SimulationOptions(CESE.getEnvironment());
		home = new UserHome(this);
		builder = new RecordBuilder();
		recentParameters = new RecentFileMenu(OpenRecentParamsAction.class);

		this.setLayout(new javax.swing.BoxLayout(this, 0));

		modelSplitPane = new javax.swing.JSplitPane();
		modelSplitPane.setOrientation(javax.swing.SwingConstants.HORIZONTAL);
		modelSplitPane.setContinuousLayout(true);
		modelSplitPane.setOneTouchExpandable(true);

		paramsPanel = new javax.swing.JPanel();
		paramsPanel.setLayout(new java.awt.BorderLayout());
		CSH.setHelpIDString(paramsPanel, "Visual.Parameters");
		modelSplitPane.setLeftComponent(paramsPanel);

		exportPane = new ExportPane();
		CSH.setHelpIDString(exportPane, "Visual.Display");
		modelSplitPane.setRightComponent(exportPane);

		this.add(modelSplitPane);

		modelInfo = new ModelInfo();

		propertyPane = new PropertyPane();
		CSH.setHelpIDString(propertyPane, "Visual.Parameters");

		paramsTabs = new JTabbedPane();
		paramsTabs.addTab(ResourceManager.getResource("Default"), propertyPane);
		paramsPanel.add(paramsTabs, "Center");

		modelSelector = new javax.swing.JComboBox();
		modelSelector.setRenderer(new ModelBoxRenderer(false));
		CSH.setHelpIDString(modelSelector, "Visual.Selector");
		modelSelector.getModel().addListDataListener(
				new javax.swing.event.ListDataListener() {

					public void contentsChanged(
							javax.swing.event.ListDataEvent evt) {
					}

					public void intervalAdded(
							javax.swing.event.ListDataEvent evt) {
						ComboBoxModel cb = (ComboBoxModel) evt.getSource();
						if (cb.getSize() > 0) {
							setModelActions(true);
						} else {
							setModelActions(false);
						}
					}

					public void intervalRemoved(
							javax.swing.event.ListDataEvent evt) {
						ComboBoxModel cb = (ComboBoxModel) evt.getSource();
						if (cb.getSize() > 0) {
							setModelActions(true);
						} else {
							setModelActions(false);
						}
					}
				});

		modelSelector.addActionListener(new java.awt.event.ActionListener() {

			public void actionPerformed(java.awt.event.ActionEvent evt) {
				JComboBox cb = (JComboBox) evt.getSource();
				if (modelInfo != null) {
					modelInfo.getHistory().setTreeSelection(
							getExportTreeSelection());
				}

				modelInfo = (ModelInfo) cb.getSelectedItem();
				if (cb.getItemCount() > 0)
					changeModel();
			}
		});

		paramsPanel.add(modelSelector, "North");

		Box statusPanel = Box.createHorizontalBox();
		statusBar = new javax.swing.JTextField(20);
		CSH.setHelpIDString(statusBar, "Visual.Status");
		statusBar.setEditable(false);
		statusPanel.add(statusBar);
		statusPanel.add(Box.createHorizontalGlue());
		dm.getDesktop().getTasks().setFont(statusBar.getFont());
		statusPanel.add(dm.getDesktop().getTasks());
		paramsPanel.add(statusPanel, "South");

		modelSplitPane.setDividerLocation(350);

		setModelActions(false);
		setRecordActions(false);
		setHistoryActions(false);

		setStatusBar();
	}

	/**
	 * Creates new model
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public void newModel() throws IOException, ClassNotFoundException {
		modelInfo.newModel();
		modelInfo.getHistory().setTreeSelection(getExportTreeSelection());

		modelInfo.setModified(false);
		changeModel();
		propertyPane.clearHighlights();
	}

	/**
	 * Changes current model via Selector. This causes restoration of model
	 * history, and selection state of the Export Tree.
	 */
	public void changeModel() {
		try {
			updateTab();

			propertyPane.setModelInfo(modelInfo);
			exportPane.setModelInfo(modelInfo);
			exportPane.setExportTreeSelection(modelInfo.getHistory()
					.getTreeSelection());

			dm.setTitle(ResourceManager.getResource("Title") + " : "
					+ modelInfo.getDisplayName());

			setHistoryActions(true);
		} catch (Exception e) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Error_initializing_model."), e);
		}
	}

	/**
	 * Installs model from jar
	 * 
	 * @param f
	 *            jar file containing models
	 * @throws IOException
	 */
	public ModelInstaller installModel(final File f) throws IOException {
		final ModelInstaller modelInstaller = new ModelInstaller();
		modelInstaller.addJarFile(f.getPath());

		java.awt.EventQueue.invokeLater(new Runnable() {
			public void run() {

				int installed = modelInstaller.getInstalledModels().size();
				if (installed > 0) {
					for (int i = 0; i < installed; i++) {
						modelSelector.addItem(modelInstaller
								.getInstalledModels().get(i));
						((ModelInfo) modelInstaller.getInstalledModels().get(i))
								.addPropertyChangeListener(ModelManager.this);
					}
				}
			}
		});

		return modelInstaller;
	}

	/**
	 * Loads model from serialized form
	 * 
	 * @param lm
	 *            model to load
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * 
	 * @return true if model parameters are loaded
	 */
	public boolean loadModel(Object lm) throws IOException,
			ClassNotFoundException {
		ModelInfo mm = new ModelInfo(lm.getClass());

		for (int i = 0; i < modelSelector.getItemCount(); i++) {
			ModelInfo mi = (ModelInfo) modelSelector.getItemAt(i);
			if (mi.getName().equals(mm.getName())) {
				if (i != modelSelector.getSelectedIndex()) {
					modelSelector.setSelectedIndex(i);
				}

				modelInfo.setModel(lm);
				propertyPane.clearHighlights();

				return true;
			}
		}
		return false;
	}

	/**
	 * Returns current model info
	 * 
	 * @return current model info
	 */
	public ModelInfo getModelInfo() {
		return modelInfo;
	}

	/**
	 * Returns model selector
	 * 
	 * @return model selector
	 */
	public JComboBox getSelector() {
		return modelSelector;
	}

	/**
	 * Returns user home
	 * 
	 * @return user home
	 */
	public UserHome getHome() {
		return home;
	}

	/**
	 * Returns simulation options
	 * 
	 * @return simulation options
	 */
	public SimulationOptions getOptions() {
		return options;
	}

	/**
	 * Returns record builder
	 * 
	 * @return record builder
	 */
	public RecordBuilder getRecordBuilder() {
		return builder;
	}

	/**
	 * Starts simulation
	 */
	public void startSimulation() {
		Object expvar[];
		TreePath exptree[];
		int sel = 0;

		propertyPane.clearHighlights();

		exptree = getExportTreeSelection();

		if (exptree != null) {
			for (int i = 0; i < exptree.length; i++) {
				expvar = exptree[i].getPath();
				for (int j = 0; j < expvar.length; j++) {
					if (expvar[j] instanceof PropertyDescriptor)
						sel++;
				}
			}
		}

		if ((sel == 0) || (exptree == null)) {
			int n = JOptionPane
					.showConfirmDialog(
							CESE.getEnvironment(),
							ResourceManager
									.getResource("Display_selection_is_empty._Continue_with_simulation?"),
							ResourceManager.getResource("Selection_is_empty"),
							JOptionPane.YES_NO_OPTION);
			if (n == JOptionPane.NO_OPTION)
				return;
		}

		setSimulationActions(true);

		worker = new SwingWorker() {
			long total = System.currentTimeMillis();

			public Object construct() {
				String titr = "";
				int count = 0;
				iteration = "";

				int itr = options.getIterations();
				if (options.isSingle())
					itr = 1;

				for (int i = 0; i < itr; i++) {
					try {
						time = System.currentTimeMillis();
						count++;

						if (!options.isSingle()) {
							if (options.isInfinite()) {
								i = itr - 2;
								titr = "/" + '\u221E';
							} else {
								titr = "/" + itr;
							}

							if (count > 0)
								iteration = " (" + count + titr + ")";
						}

						if (Thread.currentThread().isInterrupted()) {
							throw new InterruptedException();
						}

						getModelInfo().getModel().simulate();

						java.awt.EventQueue.invokeLater(new Runnable() {
							public void run() {
								String status = ResourceManager
										.getResource("Simulation_")
										+ (System.currentTimeMillis() - time)
										+ iteration;

								if (iteration.length() > 0)
									statusBar.setText(status);
							}
						});

					} catch (InterruptedException e) {
						return ResourceManager.getResource("Interrupted");
					}
				}

				return ResourceManager.getResource("Done");
			}

			public void finished() {
				statusBar.setText(ResourceManager.getResource("Elapsed_")
						+ (System.currentTimeMillis() - total));

				dm.getDesktop().getTasks().removeTask(
						ResourceManager.getResource("Performing_simulation"));
				setSimulationActions(false);
				propertyPane.updateTable(options.isHighlight());

				if (options.isHistory()) {
					modelInfo.getHistory().addHistory();
					setHistoryActions(true);
				}

				if (options.isBeep())
					Toolkit.getDefaultToolkit().beep();

				modelInfo.setModified(true);
			}

		};
		dm.getDesktop().getTasks().addTask(
				ResourceManager.getResource("Performing_simulation"));
		worker.setPriority(options.getPriority());
		worker.start();
	}

	/**
	 * Stops simulation
	 */
	public void stopSimulation() {
		worker.interrupt();
		worker = null;

		dm.getDesktop().getTasks().removeTask(
				ResourceManager.getResource("Performing_simulation"));
		setSimulationActions(false);
		propertyPane.updateTable(false);

		modelInfo.setModified(true);
	}

	/**
	 * Sets statusBar string
	 */
	public void setStatusBar() {
		String s = "";
		switch (getExpType()) {
		case 0:
			s = ResourceManager.getResource("Track_experiment");
			break;
		case 1:
			s = ResourceManager.getResource("Clamp_experiment");
			break;
		default:
			break;
		}
		if (builder.getRecFile() != null)
			s = s + " : " + builder.getRecFile().getPath();
		statusBar.setText(s);
	}

	/**
	 * Returns experiment type changer
	 * 
	 * @return experiment type changer
	 */
	public ExperimentTypeChanger getExperimentTypeChanger() {
		if (experimentType == null)
			experimentType = new ExperimentTypeChanger();
		return experimentType;
	}

	/**
	 * Returns export tree selection
	 * 
	 * @return selection in the export tree
	 */
	public TreePath[] getExportTreeSelection() {
		return exportPane.getExportTreeSelection();
	}

	/**
	 * Returns property table clamp values
	 * 
	 * @return list of clamped variables
	 */
	public Vector getClampProperties() {
		return propertyPane.getTableModel().getClampedPropertyDescriptors();
	}

	/**
	 * Sets property table clamp values
	 * 
	 * @param names
	 *            list of clamped variable names
	 */
	public void setClampProperties(Vector names) {
		propertyPane.getTableModel().setClampedPropertyDescriptors(names);
	}

	/**
	 * Exports the content of <code>PropertyTable</code> to ASCII file
	 * 
	 * @param fileName
	 *            name of the file to export data
	 * @throws IOException
	 *             if i/o problem occurs
	 */
	public synchronized void exportParametersASCII(String fileName)
			throws java.io.IOException {
		java.io.PrintWriter datExp = new java.io.PrintWriter(
				new java.io.FileWriter(fileName));

		datExp.print(propertyPane.getTableModel().exportParametersASCII());

		datExp.flush();
		datExp.close();
	}

	/**
	 * Returns type of experiment
	 * 
	 * @return type of the experiment
	 */
	public int getExpType() {
		return getExperimentTypeChanger().getExpType();
	}

	/**
	 * Sets type of experiment to clamp
	 */
	public void setClamp() {
		getExperimentTypeChanger().setClamp();
	}

	/**
	 * Sets state for model actions
	 * 
	 * @param state
	 *            state of the actions
	 */
	public void setModelActions(boolean state) {
		this.firePropertyChange("model_control", !state, state);
	}

	private void setSimulationActions(boolean state) {
		this.firePropertyChange("simulation_control", !state, state);
	}

	/**
	 * Sets state for record actions
	 * 
	 * @param state
	 *            state of the actions
	 */
	public void setRecordActions(boolean state) {
		if (builder.getRecFile() == null)
			state = false;
		this.firePropertyChange("record_control", !state, state);
		builder.getRecNumberField().setEditable(state);
	}

	/**
	 * Sets state for lock actions
	 * 
	 * @param state
	 *            state of the actions
	 */
	public void setLockActions(boolean state) {
		if (builder.getRecFile() == null)
			state = false;
		this.firePropertyChange("lock_control", !state, state);
	}

	/**
	 * Sets state for history actions
	 * 
	 * @param state
	 *            state of the actions
	 */
	public void setHistoryActions(boolean state) {
		this.firePropertyChange("history_control", !state, state);
	}

	/**
	 * Returns recent parameters submenu
	 * 
	 * @return the recentParameters
	 */
	public RecentFileMenu getRecentParameters() {
		return recentParameters;
	}

	public void propertyChange(java.beans.PropertyChangeEvent evt) {
		if (evt.getPropertyName().equals("modified"))
			updateTab();

	}

	private void updateTab() {
		paramsTabs.setTitleAt(0, modelInfo.getHistory().getParametersName());

		if (modelInfo.isModified()) {
			paramsTabs.setIconAt(0, new ImageIcon(this.getClass().getResource(
					ResourceManager.getResource("save_params_image"))));
		} else {
			paramsTabs.setIconAt(0, null);
		}
	}

	private class ExperimentTypeChanger extends JMenu implements ItemListener,
			java.beans.PropertyChangeListener {

		private int expType;

		JRadioButtonMenuItem TrackItem;

		JRadioButtonMenuItem ClampItem;

		/**
		 * Default constructor
		 */
		public ExperimentTypeChanger() {
			setText(ResourceManager.getResource("Experiment_Type"));
			setMnemonic('E');

			ButtonGroup ExpTypeGroup = new ButtonGroup();

			TrackItem = new JRadioButtonMenuItem();
			TrackItem.setText(ResourceManager.getResource("Track"));
			TrackItem.setSelected(true);
			TrackItem.setMnemonic('T');
			CSH.setHelpIDString(TrackItem, "Commands.Track_Experiment");
			TrackItem.addItemListener(this);
			ExpTypeGroup.add(TrackItem);
			add(TrackItem);

			ClampItem = new JRadioButtonMenuItem();
			ClampItem.setText(ResourceManager.getResource("Clamp"));
			ClampItem.setSelected(false);
			ClampItem.setMnemonic('C');
			CSH.setHelpIDString(ClampItem, "Commands.Clamp_Experiment");
			ClampItem.addItemListener(this);
			ExpTypeGroup.add(ClampItem);
			add(ClampItem);

			this.addPropertyChangeListener(this);
		}

		/**
		 * Switches to the clamp mode
		 */
		public void setClamp() {
			ClampItem.setSelected(true);
		}

		/**
		 * Returns experiment type
		 * 
		 * @return experiment type
		 */
		public int getExpType() {
			return expType;
		}

		/**
		 * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
		 */
		public void itemStateChanged(ItemEvent e) {
			try {
				JRadioButtonMenuItem rb = (JRadioButtonMenuItem) e.getSource();
				if (rb.isSelected()) {
					String selected = rb.getText();
					if (selected.equals(ResourceManager.getResource("Track"))) {
						expType = 0;
					} else if (selected.equals(ResourceManager
							.getResource("Clamp"))) {
						expType = 1;
					}
					setStatusBar();
				}
			} catch (Exception ex) {
			}
		}

		/**
		 * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
		 */
		public void propertyChange(java.beans.PropertyChangeEvent evt) {
			if (evt.getPropertyName().equals("record_control"))
				this.setEnabled(((Boolean) evt.getNewValue()).booleanValue());
		}
	}

}