/*
 * com.simulogic.cese -- Cell Electrophysiology Simulation Environment (CESE)
 *
 * Copyright (C) 1999 - 2005 Sergey Missan
 * 
 * Dalhousie Univ., Halifax, NS, Canada
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.simulogic.cese;

import java.awt.Toolkit;
import java.beans.Beans;
import java.beans.PropertyDescriptor;
import java.io.File;
import java.lang.reflect.Method;
import java.util.Vector;

import javax.swing.JPanel;
import javax.swing.JTextField;

import org.apache.xerces.parsers.SAXParser;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import com.simulogic.cese.actions.OpenRecentRecAction;
import com.simulogic.cese.signals.AbstractSignal;
import com.simulogic.cese.signals.WaveformSignal;
import com.javable.cese.templates.AbstractAPModel;
import com.simulogic.dataview.DataChannel;
import com.simulogic.dataview.DataFrame;
import com.simulogic.dataview.DataGroup;
import com.simulogic.dataview.TimeChannel;
import com.simulogic.utils.ExceptionDialog;
import com.simulogic.utils.IntegerField;
import com.simulogic.utils.RecentFileMenu;
import com.simulogic.utils.TYParse;

/**
 * This class creates a clamping command using XML-based command files for
 * records and signals.
 */
public class RecordBuilder extends JPanel {

	private static final String signalPackage = "com.simulogic.cese.signals.";

	private AbstractAPModel model;

	private Method setters[];

	private DataChannel command;

	private Vector record;

	private double stepsize;

	private int size;

	private int maxsize;

	private int recid;

	private int count;

	private TYParse nParser;

	private SAXParser recParser = new SAXParser();

	private SAXParser sigParser = new SAXParser();

	private IntegerField recNumberField;

	private RecentFileMenu recentRecords;

	private File recFile;

	/** Default constructor */
	public RecordBuilder() {
		recentRecords = new RecentFileMenu(OpenRecentRecAction.class);
		recNumberField = new IntegerField(0, 2);
		recNumberField.setPositive(true);
		recNumberField.setMaxConstraint(0);
		recNumberField.setEditable(false);
		recNumberField.setToolTipText(ResourceManager
				.getResource("Current_record_number"));
	}

	/**
	 * Returns recent parameters submenu
	 * 
	 * @return the recentParameters
	 */
	public RecentFileMenu getRecentRecords() {
		return recentRecords;
	}

	/**
	 * Returns record number field
	 * 
	 * @return record number field
	 */
	public JTextField getRecNumberField() {
		return recNumberField;
	}

	/**
	 * Returns record number
	 * 
	 * @return current record number
	 */
	public int getRecNumber() {
		return recNumberField.getValue();
	}

	/**
	 * Changes current record number by a specific factor
	 * 
	 * @param factor
	 *            step for changing the record number (may be negative)
	 */
	public void changeRecNumber(int factor) {
		if (recNumberField.isEditable()) {
			recNumberField.setValue(recNumberField.getValue() + factor);
		recNumberField.setToolTipText(ResourceManager
				.getResource("Current_record_number")
				+ " ("
				+ getRecNumber()
				+ "/"
				+ recNumberField.getMaxConstraint() + ")");
		CESE.getEnvironment().getModelManager().setRecordActions(true);
		}
	}

	/**
	 * Sets maximum record number based on the number of groups in the record
	 * file
	 * 
	 * @param max
	 *            maximum record number in the current record file
	 */
	public void setMaxRecNumber(int max) {
		recNumberField.setMaxConstraint(max);
	}

	/**
	 * Returns maximum record number based on the number of groups in the record
	 * file
	 * 
	 * @return maximum record number in the current record file
	 */
	public int getMaxRecNumber() {
		return recNumberField.getMaxConstraint();
	}

	/**
	 * Locks/unlocks record field
	 */
	public void changeRecLock() {
		recNumberField.setEditable(!recNumberField.isEditable());
		CESE.getEnvironment().getModelManager().setLockActions(
				recNumberField.isEditable());
	}

	/**
	 * Returns record file
	 * 
	 * @return record file describing the clamp command
	 */
	public File getRecFile() {
		return recFile;
	}

	/**
	 * Sets new record file
	 * 
	 * @param f
	 *            record file
	 */
	public void setRecFile(File f) {
		recFile = f;
		CESE.getEnvironment().getModelManager().setStatusBar();
		recNumberField.setValue(0);

		CESE.getEnvironment().getModelManager().setClamp();
		setupRecord();
		recNumberField.setToolTipText(ResourceManager
				.getResource("Current_record_number")
				+ " ("
				+ getRecNumber()
				+ "/"
				+ recNumberField.getMaxConstraint() + ")");
		CESE.getEnvironment().getModelManager().setRecordActions(true);
	}

	/**
	 * Determines the number of records in the file and clamp variable names in
	 * group 0
	 */
	public void setupRecord() {
		recid = getRecNumber();
		setRecord();
	}

	/**
	 * Creates a clamping command for a given model
	 * 
	 * @param m
	 *            model
	 */
	public void buildRecord(AbstractAPModel m) {
		model = m;
		stepsize = model.getdt();
		size = model.getsteps();

		recid = getRecNumber();
		readRecord();
		changeRecNumber(1);

		Vector c = CESE.getEnvironment().getModelManager().getClampProperties();
		setters = new Method[c.size()];
		try {
			for (int i = 0; i < c.size(); i++) {
				PropertyDescriptor pd = (PropertyDescriptor) c.get(i);
				if (pd.getWriteMethod() != null) {
					setters[i] = pd.getWriteMethod();
				}
			}
		} catch (Exception e) {
		}
	}

	/**
	 * Testing record without actual modification of models parameters
	 */
	public void testRecord() {
		stepsize = 0.05;
		size = 1;
		recid = 0;

		com.simulogic.utils.SwingWorker worker = new com.simulogic.utils.SwingWorker() {

			DataFrame frame = null;
			DataGroup group = null;

			public Object construct() {
				try {
					frame = CESE.getEnvironment().getDesktop().addFrame(
							com.simulogic.cese.ResourceManager
									.getResource("Test_record"));

					for (recid = 0; recid < getMaxRecNumber() + 1; recid++) {
						readRecord();
						size = maxsize;
						readRecord();

						if (record.size() > 0) {
							group = new DataGroup(ResourceManager
									.getResource("Record")
									+ " " + +recid, "");

							group.addChannel(new TimeChannel(stepsize, size));
							for (int i = 0; i < record.size(); i++) {
								group.addChannel((DataChannel) record.get(i));
							}
							frame.finishLoad(group);
						}
					}
				} catch (Exception e) {
				}
				return ResourceManager.getResource("Done");
			}

			public void finished() {
				CESE.getEnvironment().getDesktop().getTasks().removeTask(
						ResourceManager.getResource("Test_record"));
				Toolkit.getDefaultToolkit().beep();
			}
		};

		CESE.getEnvironment().getDesktop().getTasks().addTask(
				ResourceManager.getResource("Test_record"));
		worker.start();
	}

	/**
	 * Clamps models parameters according to the current command
	 * 
	 * @param incr
	 *            current simulation increment
	 */
	public final void clampProperties(int incr) {
		try {
			for (int i = 0; i < setters.length; i++) {
				for (int j = 0; j < record.size(); j++) {

					if (setters[i].getName().substring(3).equals(
							((DataChannel) record.get(j)).getName())) {
						Object val = new Double(((DataChannel) record.get(j))
								.getData(incr));
						Object args[] = { val };
						setters[i].invoke(model, args);
					}
				}
			}
		} catch (Exception e) {
		}
	}

	private void setRecord() {
		try {
			ClampSetter rHandler = new ClampSetter();
			recParser.setContentHandler(rHandler);
			recParser.setErrorHandler(rHandler);
			recParser.parse(getRecFile().getAbsolutePath());

			setMaxRecNumber(rHandler.getGroupCount() - 1);
			CESE.getEnvironment().getModelManager().setClampProperties(
					rHandler.getClampNames());

		} catch (Exception e) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Error_opening_record_file."), e);
		}
	}

	private void readRecord() {
		try {
			record = new Vector();
			maxsize = 0;

			RecordHandler rHandler = new RecordHandler();
			recParser.setContentHandler(rHandler);
			recParser.setErrorHandler(rHandler);
			recParser.parse(getRecFile().getAbsolutePath());
		} catch (Exception e) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Error_opening_record_file."), e);
		}
	}

	private void readSignal(String signame, String varname) {
		try {
			command = new DataChannel(varname, size);
			record.add(command);

			count = 0;

			String filename = "";
			File sDir = new File(UserHome.CESE_HOME + UserHome.FS
					+ ResourceManager.getResource("recs_dir"));
			File sFile = new File(signame);
			if (!sFile.isAbsolute()) {
				filename = sDir.getPath() + File.separatorChar
						+ sFile.getPath();
			} else {
				filename = sFile.getPath();
			}

			SignalHandler sHandler = new SignalHandler();
			sigParser.setContentHandler(sHandler);
			sigParser.setErrorHandler(sHandler);
			sigParser.parse(filename);
		} catch (Exception e) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Error_opening_signal_file."), e);
		}
	}

	private class ClampSetter extends DefaultHandler {

		private int group = 0;

		private String groupTag = "group";

		private Vector clamps = new Vector();

		private String commandTag = "command";

		private String variableAttribute = "variable";

		public void startElement(String uri, String localName, String qName,
				Attributes attributes) {
			if (group == recid) {
				if (localName.equals(commandTag)) {
					clamps.add(attributes.getValue(variableAttribute));
					nParser = new TYParse();
				}
			}
		}

		public void endElement(String uri, String localName, String qName) {
			if (localName.equals(groupTag))
				group++;
		}

		/**
		 * Number of groups in this record file
		 * 
		 * @return number of groups
		 */
		public int getGroupCount() {
			return group;
		}

		/**
		 * Names of clamp variables
		 * 
		 * @return clamp variables names
		 */
		public Vector getClampNames() {
			return clamps;
		}

		public void warning(SAXParseException ex) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Warning_parsing_record_file."), ex);
		}

		public void error(SAXParseException ex) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Error_parsing_record_file."), ex);
		}

		public void fatalError(SAXParseException ex) throws SAXException {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Fatal_error_parsing_record_file."), ex);
			throw ex;
		}
	}

	private class RecordHandler extends DefaultHandler {

		private int group = 0;

		private String formula, signalName, variableName;

		private String commandTag = "command";

		private String groupTag = "group";

		private String expressionTag = "expr";

		private String sourceAttribute = "source";

		private String variableAttribute = "variable";

		public void characters(char ch[], int start, int length) {
			String s = new String(ch, start, length);
			if (s.trim() != "")
				formula = s;
		}

		public void startElement(String uri, String localName, String qName,
				Attributes attributes) {
			if (group == recid) {
				if (localName.equals(commandTag)) {
					signalName = attributes.getValue(sourceAttribute);
					variableName = attributes.getValue(variableAttribute);
					nParser = new TYParse();
					nParser.getValue("record=" + recid);
				}
			}
		}

		public void endElement(String uri, String localName, String qName) {
			if (group == recid) {
				if ((localName.equals(expressionTag)) && (nParser != null))
					nParser.getValue(formula);
				if (localName.equals(commandTag))
					readSignal(signalName, variableName);
			}
			if (localName.equals(groupTag))
				group++;
		}

		public void warning(SAXParseException ex) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Warning_parsing_record_file."), ex);
		}

		public void error(SAXParseException ex) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Error_parsing_record_file."), ex);
		}

		public void fatalError(SAXParseException ex) throws SAXException {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Fatal_error_parsing_record_file."), ex);
			throw ex;
		}
	}

	private class SignalHandler extends DefaultHandler {

		private AbstractSignal signal;

		private double val;

		private String formula, signalName;

		private String signalTag = "signal";

		private String expressionTag = "expr";

		private String waveformTag = "waveform";

		private String fileAttribute = "file";

		public void characters(char ch[], int start, int length) {
			String s = new String(ch, start, length);
			if (s.trim() != "")
				formula = s;
		}

		public void startElement(String uri, String localName, String qName,
				Attributes attributes) {
			try {
				if ((!localName.equals(signalTag))
						&& (!localName.equals(expressionTag))) {
					signalName = localName;
					StringBuffer sn = new StringBuffer(localName);
					sn.setCharAt(0, Character.toUpperCase(localName.charAt(0)));
					sn.insert(0, signalPackage);
					sn.append("Signal");
					signal = (AbstractSignal) Beans.instantiate(ClassLoader
							.getSystemClassLoader(), sn.toString());
				}
				if (localName.equals(waveformTag)) {
					((WaveformSignal) signal).setWaveformFile(attributes
							.getValue(fileAttribute));
				}
			} catch (Exception e) {
				ExceptionDialog.showExceptionDialog(ResourceManager
						.getResource("Error"), ResourceManager
						.getResource("Can_not_load_signal."), e);
			}
		}

		public void endElement(String uri, String localName, String qName) {
			try {
				if ((localName.equals(expressionTag)) && (nParser != null))
					val = nParser.getValue(formula);
				if (signal != null)
					signal.addParameter(val);
				if ((localName.equals(signalName)) && (signal != null)) {
					double[] res = signal.buildSignal(stepsize);

					for (int i = 0; i < res.length; i++) {
						if (count < command.size())
							command.setData(count, res[i]);
						count++;
					}
					if (count > maxsize)
						maxsize = count;
					signal = null;
				}
			} catch (Exception e) {
				ExceptionDialog.showExceptionDialog(ResourceManager
						.getResource("Error"), ResourceManager
						.getResource("Can_not_build_signal."), e);
			}
		}

		public void warning(SAXParseException ex) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Warning_parsing_signal_file."), ex);
		}

		public void error(SAXParseException ex) {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Error_parsing_signal_file."), ex);
		}

		public void fatalError(SAXParseException ex) throws SAXException {
			ExceptionDialog.showExceptionDialog(ResourceManager
					.getResource("Error"), ResourceManager
					.getResource("Fatal_error_parsing_signal_file."), ex);
			throw ex;
		}
	}
}