/*
 * com.simulogic.cese -- Cell Electrophysiology Simulation Environment (CESE)
 * 
 * Copyright (C) 1999 - 2006 Sergey Missan
 * 
 * Dalhousie Univ., Halifax, NS, Canada
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 *  
 */

package com.simulogic.cese;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Vector;

import javax.swing.JComponent;
import javax.swing.beanbuilder.CustomObjectInputStream;
import javax.swing.beanbuilder.JarClassLoader;
import javax.swing.tree.TreePath;

/**
 * Class that controls history of simulation runs
 */
public class SimulationHistory extends JComponent {

	private Vector simHistory;

	private int history;

	private File parameters = null;

	private TreePath[] treeSelection;

	private ModelInfo info;

	/**
	 * Default constructor
	 * 
	 * @param i
	 *            model info
	 */
	public SimulationHistory(ModelInfo i) {
		info = i;

		simHistory = new Vector();
		addHistory();
	}

	/**
	 * Moves to the next step in the simulation history
	 */
	public void nextHistory() {
		ObjectInputStream in = null;

		if (simHistory != null) {
			if (history < simHistory.size())
				history++;

			try {
				ByteArrayInputStream bis = new ByteArrayInputStream(
						(byte[]) simHistory.get(history));
				in = new CustomObjectInputStream(bis, JarClassLoader
						.getJarClassLoader());

				info.setModel(in.readObject());
			} catch (ClassNotFoundException ex) {

			} catch (IOException ex) {

			} finally {
				try {
					if (in != null) {
						in.close();
					}
				} catch (IOException ex) {
				}
			}
		}
	}

	/**
	 * Moves to the previous step in the simulation history
	 */
	public void previousHistory() {
		ObjectInputStream in = null;

		if (simHistory != null) {
			if (history > 0)
				history--;

			try {
				ByteArrayInputStream bis = new ByteArrayInputStream(
						(byte[]) simHistory.get(history));
				in = new CustomObjectInputStream(bis, JarClassLoader
						.getJarClassLoader());

				info.setModel(in.readObject());
			} catch (ClassNotFoundException ex) {

			} catch (IOException ex) {

			} finally {
				try {
					if (in != null) {
						in.close();
					}
				} catch (IOException ex) {
				}
			}
		}
	}

	/**
	 * Returns maximum number of simulation history items
	 * 
	 * @return maximum number of simulation history items
	 */
	public int getMaxHistoryNumber() {
		if (simHistory != null)
			return simHistory.size();
		return 0;
	}

	/**
	 * Returns current history number
	 * 
	 * @return current history number
	 */
	public int getHistoryNumber() {
		return history;
	}

	/**
	 * Adds another item to the history collection
	 */
	public void addHistory() {
		byte[] buf = null;
		try {
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			ObjectOutputStream out = new ObjectOutputStream(bos);
			out.writeObject(info.getModel());
			out.close();

			buf = bos.toByteArray();
		} catch (IOException e) {
		}
		if (buf != null) {
			simHistory.addElement(buf);
			history = simHistory.size() - 1;
		}
	}

	/**
	 * Returns model parameters file name
	 * 
	 * @return model parameters file name
	 */
	public String getParametersName() {
		if (parameters != null)
			return parameters.getName();

		return ResourceManager.getResource("Default");
	}

	/**
	 * Returns model parameters file
	 * 
	 * @return model parameters file
	 */
	public File getParametersFile() {
		return parameters;
	}

	/**
	 * Sets model parameters file
	 * 
	 * @param params
	 *            model parameters file
	 */
	public void setParametersFile(File params) {
		this.parameters = params;
	}

	/**
	 * Returns display tree selection
	 * 
	 * @return display tree selection
	 */
	public TreePath[] getTreeSelection() {
		return treeSelection;
	}

	/**
	 * Sets display tree selection
	 * 
	 * @param treeSelection
	 *            display tree selection
	 */
	public void setTreeSelection(TreePath[] treeSelection) {
		this.treeSelection = treeSelection;
	}

}
