/*
 * com.simulogic.cese -- Cell Electrophysiology Simulation Environment (CESE)
 * 
 * Copyright (C) 1999 - 2005 Sergey Missan
 * 
 * Dalhousie Univ., Halifax, NS, Canada
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 *  
 */

package com.simulogic.cese;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.swing.filechooser.FileSystemView;

import com.simulogic.utils.ExceptionDialog;

/**
 * This class handles creation and population of the user home parameters
 * directory
 */
public class UserHome {

	/**
	 * the file separator in a platform dependent way
	 */
	public static final String FS = System.getProperty("file.separator");

	/**
	 * major version identificator
	 */
	public static final String CESE_VERSION_MAJOR = ResourceManager
			.getResource("Version_major");

	/**
	 * major version identificator
	 */
	public static final String CESE_VERSION_MINOR = ResourceManager
			.getResource("Version_minor");

	/**
	 * name of the directory containing the setup
	 */
	public static final String CESE_HOME = FileSystemView.getFileSystemView()
			.getDefaultDirectory()
			+ FS + "cese" + "-" + CESE_VERSION_MAJOR;

	/**
	 * name of the directory containing the old setup
	 */
	public static final String OLD_CESE_HOME = System.getProperty("user.home")
			+ FS + "cese" + "-" + CESE_VERSION_MAJOR;

	private ModelManager manager;

	/**
	 * Default constructor
	 * 
	 * @param m
	 *            model manager
	 */
	public UserHome(ModelManager m) {
		this.manager = m;
	}

	/**
	 * Creates user home directory structure
	 */
	public void populateUserHome() {
		File home = new File(CESE_HOME);
		File old_home = new File(OLD_CESE_HOME);

		if (old_home.exists() && !home.exists()) {
			new File(CESE_HOME).mkdir();

			replicateDirectory(OLD_CESE_HOME + FS
					+ ResourceManager.getResource("mods_dir"), CESE_HOME + FS
					+ ResourceManager.getResource("mods_dir"));

			replicateDirectory(OLD_CESE_HOME + FS
					+ ResourceManager.getResource("cellml_dir"), CESE_HOME + FS
					+ ResourceManager.getResource("cellml_dir"));

			replicateDirectory(OLD_CESE_HOME + FS
					+ ResourceManager.getResource("recs_dir"), CESE_HOME + FS
					+ ResourceManager.getResource("recs_dir"));

			replicateDirectory(OLD_CESE_HOME + FS
					+ ResourceManager.getResource("params_dir"), CESE_HOME + FS
					+ ResourceManager.getResource("params_dir"));

			replicateDirectory(OLD_CESE_HOME + FS
					+ ResourceManager.getResource("results_dir"), CESE_HOME
					+ FS + ResourceManager.getResource("results_dir"));

			old_home.renameTo(new File(System.getProperty("user.home") + FS
					+ "old_cese" + "-" + CESE_VERSION_MAJOR));
		}

		if (!home.exists()) {
			new File(CESE_HOME).mkdir();

			new File(CESE_HOME + FS + ResourceManager.getResource("mods_dir"))
					.mkdir();

			replicateDirectory(ClassLoader.getSystemResource(
					ResourceManager.getResource("cellml_dir")).getPath(),
					CESE_HOME + FS + ResourceManager.getResource("cellml_dir"));

			replicateDirectory(ClassLoader.getSystemResource(
					ResourceManager.getResource("recs_dir")).getPath(),
					CESE_HOME + FS + ResourceManager.getResource("recs_dir"));

			replicateDirectory(ClassLoader.getSystemResource(
					ResourceManager.getResource("params_dir")).getPath(),
					CESE_HOME + FS + ResourceManager.getResource("params_dir"));

			replicateDirectory(ClassLoader.getSystemResource(
					ResourceManager.getResource("results_dir")).getPath(),
					CESE_HOME + FS + ResourceManager.getResource("results_dir"));
		}

		this.autoInstall();
	}

	/**
	 * Replicates content of the install directory into the user home
	 * 
	 * @param srcdir
	 *            source directory
	 * @param dstdir
	 *            destination directory
	 */
	private void replicateDirectory(String srcdir, String dstdir) {
		new File(dstdir).mkdir();

		File fDir = new File(java.net.URLDecoder.decode(srcdir));
		File[] files = fDir.listFiles();

		try {
			for (int i = 0; i < files.length; i++) {
				if (files[i].isFile()) {
					File dstFile = new File(dstdir + FS + files[i].getName());
					copy(files[i], dstFile);
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Copies src file to dst file
	 */
	private void copy(File src, File dst) throws IOException {
		InputStream in = new FileInputStream(src);
		OutputStream out = new FileOutputStream(dst);

		// Transfer bytes from in to out
		byte[] buf = new byte[1024];
		int len;
		while ((len = in.read(buf)) > 0) {
			out.write(buf, 0, len);
		}
		in.close();
		out.close();
	}

	/**
	 * Autoinstalls models from the default location
	 */
	private void autoInstall() {
		com.simulogic.utils.SwingWorker worker = new com.simulogic.utils.SwingWorker() {
			Exception ex = null;

			public Object construct() {
				try {
					ex = null;

					// first use global install
					File fDir = new File(java.net.URLDecoder.decode(ClassLoader
							.getSystemResource(
									ResourceManager.getResource("mods_dir"))
							.getPath()));

					File[] files = fDir.listFiles();

					if (files != null) {
						for (int i = 0; i < files.length; i++) {
							manager.installModel(files[i]);
						}
					}

					// then use local install
					fDir = new File(CESE_HOME + FS
							+ ResourceManager.getResource("mods_dir"));

					files = fDir.listFiles();

					if (files != null) {
						for (int i = 0; i < files.length; i++) {
							manager.installModel(files[i]);
						}
					}

				} catch (Exception e) {
					ex = e;
				}
				return ResourceManager.getResource("Done");
			}

			public void finished() {
				CESE.getEnvironment().getDesktop().getTasks().removeTask(
						ResourceManager.getResource("Autoinstalling_models"));

				if (ex != null)
					ExceptionDialog.showExceptionDialog(ResourceManager
							.getResource("Error"), ResourceManager
							.getResource("Error_auto-installing_models."), ex);
			}
		};

		CESE.getEnvironment().getDesktop().getTasks().addTask(
				ResourceManager.getResource("Autoinstalling_models"));
		worker.start();
	}

}
