/*
 * com.simulogic.dataview -- The Java2D Plotting Framework
 *
 * Copyright (C) 1999 - 2005 Sergey Missan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.simulogic.dataview.analysis;

import com.simulogic.dataview.DataChannel;

/**
 * This helper class improves performance of the batch statistics operations.
 * <p>
 * The performance of the statistics operations can be improved by caching some
 * frequently used values instead of recalculating them from scratch every time.
 * This class does exactly this transparently to user, it delegates (in most
 * cases) the actual calculations to <code>ChannelStats</code>.
 * <p>
 * Note, that usage of this class is advantageous only when you are performing
 * several consequitive statistics routines on the channel, if you need only one
 * or two statistics parameters, you're probably better off with a direct call
 * to <code>ChannelStats</code>.
 * <p>
 * This class is created with reuse in mind: you can reset its state by calling
 * <code>clear()</code> instead of instantiating the object every time.
 */
public class ChannelStatsProxy {

    private double n;
    private double mean;
    private double variance;
    private int negativePeak;
    private int positivePeak;
    private int absolutePeak;

    /**
     * Creates new ChannelStatsProxy
     */
    public ChannelStatsProxy() {
        clear();
    }

    /**
     * Resets class state to the initial conditions
     */
    public void clear() {
        n = Double.NaN;
        mean = Double.NaN;
        variance = Double.NaN;
        negativePeak = -1;
        positivePeak = -1;
        absolutePeak = -1;
    }

    /**
     * Returns the number of data points between xMin and xMax for the given x
     * channel / y channel pair
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return number of data points between xMin and xMax
     */
    public final double getN(double xMin, double xMax, DataChannel xChannel,
            DataChannel yChannel) {
        if (Double.isNaN(n))
                n = ChannelStats.getN(xMin, xMax, xChannel, yChannel);

        return n;
    }

    /**
     * Returns the mean of a data sequence. That is <code>sum / size</code>
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getMean(double xMin, double xMax, DataChannel xChannel,
            DataChannel yChannel) {
        if (Double.isNaN(mean))
                mean = ChannelStats.getMean(xMin, xMax, xChannel, yChannel);

        return mean;
    }

    /**
     * Returns the variance of a data sequence. That is
     * <code>(sumOfSquares - mean*sum) / size</code> with
     * <code>mean = sum/size</code>
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getVariance(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        if (Double.isNaN(variance))
                variance = ChannelStats.getVariance(xMin, xMax, xChannel,
                        yChannel);

        return variance;
    }

    /**
     * Returns the standard deviation of a data sequence. That is
     * <code>Math.sqrt(variance)</code>
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getStandardDeviation(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        return Math.sqrt(getVariance(xMin, xMax, xChannel, yChannel));
    }

    /**
     * Returns the standard error of a data sequence. That is
     * <code>Math.sqrt(variance/size)</code>
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getStandardError(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        return Math.sqrt(getVariance(xMin, xMax, xChannel, yChannel)
                / getN(xMin, xMax, xChannel, yChannel));
    }

    /**
     * Returns the sum of squared mean deviations of a data sequence. That is
     * <code>variance * (size-1) == Sum( (data[i] - mean)^2 )</code>
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getSumOfSquaredDeviations(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        return Math.sqrt(getVariance(xMin, xMax, xChannel, yChannel)
                * (getN(xMin, xMax, xChannel, yChannel) - 1));
    }

    private final int getNegativePeakIndex(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        if (negativePeak == -1)
                negativePeak = ChannelStats.getNegativePeakIndex(xMin, xMax,
                        xChannel, yChannel);

        return negativePeak;
    }

    private final int getPositivePeakIndex(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        if (positivePeak == -1)
                positivePeak = ChannelStats.getPositivePeakIndex(xMin, xMax,
                        xChannel, yChannel);

        return positivePeak;
    }

    private final int getAbsolutePeakIndex(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        if (absolutePeak == -1)
                absolutePeak = ChannelStats.getAbsolutePeakIndex(xMin, xMax,
                        xChannel, yChannel);

        return absolutePeak;
    }

    /**
     * Returns the negative peak value of a data sequence
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getNegativePeak(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        double peak = 0;

        int i = getNegativePeakIndex(xMin, xMax, xChannel, yChannel);

        if ((i < xChannel.size()) && (i >= 0)) peak = yChannel.getData(i);

        return peak;
    }

    /**
     * Returns the negative peak time of a data sequence
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getNegativePeakTime(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        double time = 0;

        int i = getNegativePeakIndex(xMin, xMax, xChannel, yChannel);

        if ((i < xChannel.size()) && (i >= 0)) time = xChannel.getData(i);

        return time;
    }

    /**
     * Returns the positive peak value of a data sequence
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getPositivePeak(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        double peak = 0;

        int i = getPositivePeakIndex(xMin, xMax, xChannel, yChannel);

        if ((i < xChannel.size()) && (i >= 0)) peak = yChannel.getData(i);

        return peak;
    }

    /**
     * Returns the positive peak time of a data sequence
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getPositivePeakTime(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        double time = 0;

        int i = getPositivePeakIndex(xMin, xMax, xChannel, yChannel);

        if ((i < xChannel.size()) && (i >= 0)) time = xChannel.getData(i);

        return time;
    }

    /**
     * Returns the absolute peak value of a data sequence
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getAbsolutePeak(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        double peak = 0;

        int i = getAbsolutePeakIndex(xMin, xMax, xChannel, yChannel);

        if ((i < xChannel.size()) && (i >= 0))
                peak = Math.abs(yChannel.getData(i));

        return peak;
    }

    /**
     * Returns the absolute peak time of a data sequence
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getAbsolutePeakTime(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        double time = 0;

        int i = getAbsolutePeakIndex(xMin, xMax, xChannel, yChannel);

        if ((i < xChannel.size()) && (i >= 0)) time = xChannel.getData(i);

        return time;
    }

    /**
     * Returns the area under the curve of a data sequence.
     * <p>
     * This method assumes that xchannel values are monotonically rising with
     * index. All values that do not satisfy <code>x[n]&lt;x[n+1]</code>
     * criterion are discarded.
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return double
     */
    public final double getCurveArea(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        return ChannelStats.getCurveArea(xMin, xMax, xChannel, yChannel);
    }

    /**
     * Returns the maximum slope <code>abs(delta(data)/delta(time))</code>
     * time
     * <p>
     * This method assumes that xchannel values are monotonically rising with
     * index. All values that do not satisfy <code>x[n]&lt;x[n+1]</code>
     * criterion are discarded.
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param direction search direction (to left or to right from the peak)
     * @param xChannel x channel
     * @param yChannel y channel
     * @return int
     */
    public final double getMaxSlopeTime(double xMin, double xMax,
            int direction, DataChannel xChannel, DataChannel yChannel) {
        double time = 0;

        int i = getAbsolutePeakIndex(xMin, xMax, xChannel, yChannel);
        int j = ChannelStats.getMaxSlopeIndex(xMin, xMax, i, direction,
                xChannel, yChannel);

        if ((j < xChannel.size()) && (j >= 0)) time = xChannel.getData(j);

        return time;
    }

    /**
     * Returns the maximum slope <code>abs(delta(data)/delta(time))</code>
     * value
     * <p>
     * This method assumes that xchannel values are monotonically rising with
     * index. All values that do not satisfy <code>x[n]&lt;x[n+1]</code>
     * criterion are discarded.
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param direction search direction (to left or to right from the p)
     * @param xChannel x channel
     * @param yChannel y channel
     * @return int
     */
    public final double getMaxSlope(double xMin, double xMax, int direction,
            DataChannel xChannel, DataChannel yChannel) {
        double slope = 0;
        double delta_m = 0;
        double delta_p = 0;
        int s = -1;

        int i = getAbsolutePeakIndex(xMin, xMax, xChannel, yChannel);
        s = ChannelStats.getMaxSlopeIndex(xMin, xMax, i, direction, xChannel,
                yChannel);

        if ((s < xChannel.size()) && (s >= 0)) {
            if (s > 0)
                    delta_m = Math.abs((yChannel.getData(s) - yChannel
                            .getData(s - 1))
                            / (xChannel.getData(s) - xChannel.getData(s - 1)));

            if ((s < xChannel.size() - 1) && (s < yChannel.size() - 1))
                    delta_p = Math.abs((yChannel.getData(s) - yChannel
                            .getData(s + 1))
                            / (xChannel.getData(s) - xChannel.getData(s + 1)));
        }

        if (delta_p > delta_m) {
            slope = delta_p;
        } else {
            slope = delta_m;
        }
        return slope;
    }
    
    /**
     * Returns the maximum slope <code>abs(delta(data)/delta(time))</code>
     * index
     * <p>
     * 
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     * @return slope
     */
    public final double getAbsoluteSlope(double xMin, double xMax, DataChannel xChannel, DataChannel yChannel) {
    	return ChannelStats.getAbsoluteSlope(xMin, xMax, xChannel, yChannel);
    }
   

    /**
     * Returns the half-width of curve in time units
     * <p>
     * This method assumes that xchannel values are monotonically rising with
     * index. All values that do not satisfy <code>x[n]&lt;x[n+1]</code>
     * criterion are discarded.
     * 
     * @return int
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     */
    public final double getHalfWidthTime(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        int rise50 = -1;
        int decay50 = -1;
        double width = 0;

        int i = getAbsolutePeakIndex(xMin, xMax, xChannel, yChannel);
        rise50 = ChannelStats.getSlopeLevelIndex(xMin, xMax, i,
                ChannelStats.LEFT_SLOPE, 50, xChannel, yChannel);
        decay50 = ChannelStats.getSlopeLevelIndex(xMin, xMax, i,
                ChannelStats.RIGHT_SLOPE, 50, xChannel, yChannel);

        if ((rise50 < xChannel.size()) && (decay50 < xChannel.size())
                && (rise50 >= 0) && (decay50 >= 0))
                width = Math.abs(xChannel.getData(decay50)
                        - xChannel.getData(rise50));

        return width;
    }

    /**
     * Returns the time for curve rise from 10 to 90%
     * <p>
     * This method assumes that xchannel values are monotonically rising with
     * index. All values that do not satisfy <code>x[n]&lt;x[n+1]</code>
     * criterion are discarded.
     * 
     * @return int
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     */
    public final double get10_90RiseTime(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        int rise10 = -1;
        int rise90 = -1;
        double time = 0;

        int i = getAbsolutePeakIndex(xMin, xMax, xChannel, yChannel);
        rise10 = ChannelStats.getSlopeLevelIndex(xMin, xMax, i,
                ChannelStats.LEFT_SLOPE, 10, xChannel, yChannel);
        rise90 = ChannelStats.getSlopeLevelIndex(xMin, xMax, i,
                ChannelStats.LEFT_SLOPE, 90, xChannel, yChannel);

        if ((rise10 < xChannel.size()) && (rise90 < xChannel.size())
                && (rise10 >= 0) && (rise90 >= 0))
                time = Math.abs(xChannel.getData(rise10)
                        - xChannel.getData(rise90));

        return time;
    }

    /**
     * Returns the time for curve decay from 90 to 10%
     * <p>
     * This method assumes that xchannel values are monotonically rising with
     * index. All values that do not satisfy <code>x[n]&lt;x[n+1]</code>
     * criterion are discarded.
     * 
     * @return int
     * @param xMin minimal limit for x
     * @param xMax maximal limit for x
     * @param xChannel x channel
     * @param yChannel y channel
     */
    public final double get90_10DecayTime(double xMin, double xMax,
            DataChannel xChannel, DataChannel yChannel) {
        int decay10 = -1;
        int decay90 = -1;
        double time = 0;

        int i = getAbsolutePeakIndex(xMin, xMax, xChannel, yChannel);
        decay10 = ChannelStats.getSlopeLevelIndex(xMin, xMax, i,
                ChannelStats.RIGHT_SLOPE, 10, xChannel, yChannel);
        decay90 = ChannelStats.getSlopeLevelIndex(xMin, xMax, i,
                ChannelStats.RIGHT_SLOPE, 90, xChannel, yChannel);

        if ((decay10 < xChannel.size()) && (decay90 < xChannel.size())
                && (decay10 >= 0) && (decay90 >= 0))
                time = Math.abs(xChannel.getData(decay90)
                        - xChannel.getData(decay10));

        return time;
    }

}