/*
 * com.simulogic.dataview -- The Java2D Plotting Framework
 *
 * Copyright (C) 1999 - 2005 Sergey Missan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.simulogic.dataview.export;

import java.io.IOException;

import ucar.ma2.ArrayAbstract;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriteable;

import com.simulogic.dataview.DataChannel;
import com.simulogic.dataview.DataGroup;
import com.simulogic.dataview.DataStorage;
import com.simulogic.dataview.ResourceManager;
import com.simulogic.dataview.ViewContent;

/**
 * Class that exports data to NetCDF format
 */
public class NetcdfExport {

	private com.simulogic.dataview.ViewContent view;

	/**
	 * Default constructor
	 * 
	 * @param v
	 *            view content
	 */
	public NetcdfExport(ViewContent v) {
		view = v;
	}

	/**
	 * Exports the content of <code>DataStorage</code> to NetCDF file
	 * 
	 * @param fileName
	 *            name of the file to export data
	 * @throws IOException
	 *             if i/o problem occurs
	 */
	public synchronized void export(String fileName) throws java.io.IOException {
		NetcdfFileWriteable datExp = new NetcdfFileWriteable();

		datExp.setName(fileName);
		datExp.addGlobalAttribute("title", view.getComment());

		if ((view.getTabs().indexOfTab(ResourceManager.getResource("Cursors")) == view
				.getTabs().getSelectedIndex())
				&& (view.getTabs().indexOfTab(
						ResourceManager.getResource("Cursors")) != -1)) {
			exportDataMarray(datExp, view.getView().getCursors()
					.getCursorsStorage());
		} else {
			exportDataMarray(datExp, view.getView().getStorage());
		}
		datExp.flush();
		datExp.close();
	}

	/**
	 * Exports data to multidimensional array.
	 * 
	 * Only channels with attribute NORMAL_CHANNEL (default) will be exported
	 * 
	 * @param datExp
	 *            netCDF writtable file
	 * @param storage
	 *            data storage to export
	 * @throws java.io.IOException
	 *             if i/o problem occurs
	 */
	public void exportDataMarray(NetcdfFileWriteable datExp, DataStorage storage)
			throws java.io.IOException {

		DataGroup group = null;
		DataChannel channel = null;

		Dimension xdim = datExp.addDimension("X", -1);

		for (int i = 0; i < storage.getGroupsSize(); i++) {
			group = storage.getGroup(i);
			DataChannel xchannel = storage.getChannel(i, group.getXChannel());
			String var = "chn_" + i + "_x";

			datExp.addVariable(var, double.class, new Dimension[] { xdim });
			datExp.addVariableAttribute(var, "long_name", xchannel.getName());
			datExp.addVariableAttribute(var, "units", xchannel.getUnits());
			for (int j = 0; j < storage.getChannelsSize(i); j++) {
				channel = storage.getChannel(i, j);
				var = "chn_" + i + "_" + j + "_data";

				if (channel.getAttribute().isNormal() && (channel != xchannel)) {
					datExp.addVariable(var, double.class,
							new Dimension[] { xdim });
					datExp.addVariableAttribute(var, "long_name", channel
							.getName());
					datExp.addVariableAttribute(var, "units", channel
							.getUnits());
				}
			}
		}

		datExp.create();

		for (int i = 0; i < storage.getGroupsSize(); i++) {
			group = storage.getGroup(i);
			DataChannel xchannel = storage.getChannel(i, group.getXChannel());
			String var = "chn_" + i + "_x";
			int size = xchannel.size();
			double[] data = new double[size];
			for (int k = 0; k < size; k++) {
				data[k] = xchannel.getData(k);
			}
			datExp.write(var, ArrayAbstract.factory(data));
			for (int j = 0; j < storage.getChannelsSize(i); j++) {
				channel = storage.getChannel(i, j);
				var = "chn_" + i + "_" + j + "_data";

				if (channel.getAttribute().isNormal() && (channel != xchannel)) {
					for (int k = 0; k < size; k++) {
						data[k] = channel.getData(k);
					}
					datExp.write(var, ArrayAbstract.factory(data));
				}
			}
		}

	}

}
